"""One Population Key Driver Analysis Class (NWPopDelta1/WPopDelta1)"""
import numpy as np

from .format_ouput import format_nwpopdelta1, format_wpopdelta1
from .miner import Miner


class PopDelta1:
    """"
    The operator implementing PopDelta1 (PopDelta 1 population).

    This operator works both without (NWPopDelta1) and with a target (WPopDelta1)
    column, where NW=non-weighted (without target) and W=weighted (with
    target)

    1) NWPopDelta1 reduces to a frequent itemset operator, where the counts
    are the number of rows an itemset is found in the input dataframe.
    Additional metrics are computed once the output is formatted.

    2) WPopDelta1 is a variation of frequent itemset operator, where the
    counts are the number of rows an itemset is found in the input
    dataframe weighted by the target column. As in the case of NWPopDelta1,
    additional fields are when the output is formatted.

    3) After the first exploratory batch has been used, PopDelta1 will perform
    lookups on the selected itemsets for all the sucessive batches.
    """

    def __init__(self, popdelta_status):
        self.pop_delta_status = popdelta_status
        self.target = popdelta_status.target
        self.has_target = popdelta_status.has_target
        self.weighted = self.target is not None
        self.max_len = popdelta_status.max_len
        self._initialized = False
        self.cumulative = None
        self.nw_count_left = 0
        self.w_count_left = 0

    def reset(self):
        """
        Reset Operator
        """
        self.initialized = False
        self.nw_count_left = 0
        self.w_count_left = 0

    def update_statistics(self, batch, weight_vector):
        """
        Update counters about the rows/weights that have been processed.

        Args:
            batch (pd.DataFrame): current batch
            weight_vector (np.array): current weights vector
        """

        self.nw_count_left += len(batch)
        if self.weighted:
            self.w_count_left += np.sum(weight_vector)

    def format_output(self):
        """
        Format PopDelta1 output for the front-end
        """

        if self.cumulative.empty:
            return self.cumulative

        if not self.weighted:
            return format_nwpopdelta1(df=self.cumulative,
                                 nw_sum_left=self.nw_count_left,
                                 scaling_factors=self._scaling_factors)

        return format_wpopdelta1(df=self.cumulative,
                            nw_sum_left=self.nw_count_left,
                            w_sum_left=self.w_count_left,
                            scaling_factors=self._scaling_factors)

    def process_batch(self, batches, scaling_factors):
        """
        Process Batch for PopDelta1:

        1) Separate the dataframe from the weight vector (if applies)
        2) Find optimal supports and incrementally stream the results
        3) Once the optimal support is found, mine the whole first batch
        and all the other subsequent batches from the input streams
        4) Cumulate the past results with the ones obtained from the current
        batch
        5) Format the batches for the output

        Args:
        batches ([pd.DataFrame]): list of batches (one) from the input stream
        Yields:
        output [pd.DataFrame]: results from the PopDelta1
        """

        batch = batches[0]
        miner = Miner(self.target, self.max_len)
        batch, weight_vector = miner.extract_weight_vector(batch)
        self._scaling_factors = scaling_factors

        if not self._initialized:
            for result in miner.find_support(batch, weight_vector):
                self.cumulative, self.nw_count_left, self.w_count_left = result
                output = self.format_output()
                if not output.empty:
                    yield output
            self._initialized = True
        else:
            if not self.has_target:
                weight_vector = None
            self.update_statistics(batch, weight_vector)
            batch_result = miner.lookup_itemset(itemsets=self.cumulative.index,
                                                df=batch,
                                                weight_vector=weight_vector,
                                                non_weighted=True,
                                                weighted=self.weighted)

            self.cumulative = self.cumulative.append(batch_result).groupby(level=0, sort=False).sum()
            output = self.format_output()
            if not output.empty:
                yield output
            return



"""Two Population Key Driver Analysis Class (NWPopDelta2/WPopDelta2)"""

import numpy as np
from itertools import zip_longest

from .format_ouput import format_nwpopdelta2, format_wpopdelta2
from .miner import Miner


class PopDelta2:
    """"
    Class implementing PopDelta2 (PopDelta 2 populations).

    This operator works both without (NWPopDelta2) and with a target (WPopDelta2)
    column (where NW=non-weighted (without target) and W=weighted (with
    target))

    1) NWPopDelta2 reduces to a DIFF operator, where the counts are the number
    of rows an itemset is found in the input dataframes LEFT and RIGHT.
    Given that the itemsets mined in LEFT and RIGHT are likely not
    identical we use lookups to get counts of those itemsets.

    2) WPopDelta2 is a weighted DIFF operator, where the counts are the number
    of rows an itemset is found in the input dataframea LEFT and RIGHT
    weighted by the target columns. As in the case of NWPopDelta2, additional
    fields are computed when the output is formatted. As in the previous
    case the itemsets found in LEFT and RIGHT are unlikely to be
    identical: we use lookups to get the counts for those itemsets.

    3) After the first exploratory batch has been used, PopDelta2 will perform
    lookups on the selected itemsets for all the sucessive batches.
    """

    def __init__(self, popdelta_status):
        self.popdelta_status = popdelta_status
        self.target = popdelta_status.target
        self.has_target = popdelta_status.has_target
        self.weighted = self.target is not None
        self.max_len = popdelta_status.max_len
        self._initialized = False
        self.cumulative = None

        self.nw_count_left = 0
        self.w_count_left = 0
        self.nw_count_right = 0
        self.w_count_right = 0

    def reset(self):
        self._initialized = False
        self.nw_count_left = 0
        self.w_count_left = 0
        self.nw_count_right = 0
        self.w_count_right = 0

    def _update_statistics(self):
        """
        Update counters about the rows/weights that have been processed.
        """

        self.nw_count_left += len(self.left_batch)
        self.nw_count_right += len(self.right_batch)
        if self.weighted:
            self.w_count_left += np.sum(self.left_weight_vector)
            self.w_count_right += np.sum(self.right_weight_vector)

    def _join_batch_result(self, left_result, right_result, left_miner, right_miner):
        """
        Join the results from the current batches, if it is the first
        iteration perform lookups to ensure the results are consistent across
        the two result batches.

        Args:
        left_result (pd.DataFrame): mined itemsets from the current
        left batch right_result (pd.DataFrame): mined itemsets from the
                                                current right batch

        Returns: result: joined results from current batches
        """

        if not self._initialized:

            def augment(augmentee, augmentor, miner, batch, weight_vector, result):
                """
                Lookup itemsets from missing itemsets list
                """
                missing = np.array(list(augmentor.difference(augmentee)), dtype=tuple)
                found = miner.lookup_itemset(missing, batch, weight_vector=weight_vector,
                                             non_weighted=True, weighted=self.weighted)
                augmented_result = result.append(found)
                augmented_result.index.name = "itemset"
                return augmented_result

            # compute itemsets from the two sides
            left_itemsets = set(left_result.index.values)
            right_itemsets = set(right_result.index.values)

            # lookup on a partial version of the data
            left_batch = self.left_batch[:self.nw_count_left]
            right_batch = self.right_batch[:self.nw_count_right]

            if self.has_target:
                left_weight_vector = self.left_weight_vector[:self.nw_count_left]
                right_weight_vector = self.right_weight_vector[:self.nw_count_right]
            else:
                left_weight_vector = None
                right_weight_vector = None

            left_result = augment(left_itemsets, right_itemsets, left_miner,
                                  left_batch, left_weight_vector, left_result)

            right_result = augment(right_itemsets, left_itemsets, right_miner,
                                   right_batch, right_weight_vector, right_result)

        # join results from the two sides
        result = left_result.join(right_result, how="outer", lsuffix="_left", rsuffix="_right")
        return result

    def _cumulate_results(self, result, dummy):
        if dummy is None:
            self.cumulative = self.cumulative.append(result).groupby(level=0, sort=False).sum()

        else:
            result.columns = [col_name + dummy for col_name in result.columns]
            type_cast = {col_name: int if "nw" in col_name.split("_") else float for col_name in result.columns}
            cumulated = self.cumulative[result.columns].append(result).astype(type_cast)
            self.cumulative[result.columns] = cumulated.groupby(level=0, sort=False) \
                                                       .sum() \
                                                       .set_index((self.cumulative.index))

    def format_output(self):
        """
        Format PopDelta2 output for the front-end
        """

        if self.cumulative.empty:
            return self.cumulative

        if not self.weighted:
            return format_nwpopdelta2(df=self.cumulative,
                                 nw_sum_left=self.nw_count_left,
                                 nw_sum_right=self.nw_count_right,
                                 scaling_factors=self._scaling_factors)

        return format_wpopdelta2(df=self.cumulative,
                            nw_sum_left=self.nw_count_left,
                            nw_sum_right=self.nw_count_right,
                            w_sum_left=self.w_count_left,
                            w_sum_right=self.w_count_right,
                            scaling_factors=self._scaling_factors)

    def process_batch(self, batches, scaling_factors):
        """
        Process Batch for PopDelta2:

        1) Separate the dataframe from the weight vector (if applies)
        2) Find optimal supports and incrementally stream the results
        3) Once the optimal support is found, mine the whole first batches
        and all the other subsequent batches from the input streams
        4) Cumulate the past results with the ones obtained from the current
        batch
        5) Format the batches for the output

        Args:
        batches ([pd.DataFrame]): list of batches (two) from the input stream
        Yields:
        output [pd.DataFrame]: results from the PopDelta2
        """

        left_miner = Miner(self.target, self.max_len)
        right_miner = Miner(self.target, self.max_len)
        self.left_batch, self.left_weight_vector = left_miner.extract_weight_vector(batches[0])
        self.right_batch, self.right_weight_vector = right_miner.extract_weight_vector(batches[1])
        self._scaling_factors = scaling_factors

        if not self._initialized:
            left_iterator = left_miner.find_support(self.left_batch, self.left_weight_vector)
            right_iterator = right_miner.find_support(self.right_batch, self.right_weight_vector)

            for left_iter_result, right_iter_result in zip_longest(left_iterator, right_iterator):
                if left_iter_result is not None:
                    left_result, self.nw_count_left, self.w_count_left = left_iter_result
                if right_iter_result is not None:
                    right_result, self.nw_count_right, self.w_count_right = right_iter_result

                self.cumulative = self._join_batch_result(left_result, right_result, left_miner, right_miner)
                output = self.format_output()

                if not output.empty:
                    yield output
            self._initialized = True
        else:
            self._update_statistics()
            if not self.left_batch.empty:
                left_result = left_miner.lookup_itemset(itemsets=self.cumulative.index,
                                                        df=self.left_batch,
                                                        weight_vector=self.left_weight_vector,
                                                        non_weighted=True,
                                                        weighted=self.weighted)
            if not self.right_batch.empty:
                right_result = right_miner.lookup_itemset(itemsets=self.cumulative.index,
                                                          df=self.right_batch,
                                                          weight_vector=self.right_weight_vector,
                                                          non_weighted=True,
                                                          weighted=self.weighted)

            if not self.left_batch.empty and not self.right_batch.empty:
                batch_result = self._join_batch_result(left_result, right_result, left_miner, right_miner)
                self._cumulate_results(batch_result, dummy=None)
            elif not self.left_batch.empty and self.right_batch.empty:
                self._cumulate_results(left_result, dummy="_left")
            elif self.left_batch.empty and not self.right_batch.empty:
                self._cumulate_results(right_result, dummy="_right")

            output = self.format_output()
            if not output.empty:
                yield output
            return



"""Support class to mine itemsets from dataframes using fpgrowth or lookups"""
from collections import defaultdict

from fim import fpgrowth
import pandas as pd
import numpy as np


class Miner():
    """
    Miner class that provide access to FIS mining and lookup functionalities
    """

    def __init__(self, target, max_len):
        self._target = target
        self._max_len = max_len
        self._has_target = self._target is not None

        self._max_itemest_per_batch = 1 * 10 ** 4
        self._lookup_query_step = 1000

    def extract_weight_vector(self, df):
        """
        Separates weight vector from dataframe using target column and remove it from the dataframe

        Returns:
            [df]: original dataframe without target columns
            [weight_vector]: weights_vector
        """

        if self._has_target:
            weight_vector = df[self._target].values
            df.drop(self._target, axis=1, inplace=True)
        else:
            weight_vector = np.ones(len(df), dtype=np.int_)

        if set(weight_vector) == {0}:
            raise DavosException(f'Target vector {self._target} can not be constant-zero vector.')
        return df.astype(bool), weight_vector

    def get_transactions(self, df, weight_vector):
        """
        Format hot-ecoded matrix in  transaction map {transaction : weight}
        Returns:
            [transactions]: transaction map
        """

        def to_transaction(row):
            return tuple(sorted(df.columns[row.values == 1]))

        transactions = pd.DataFrame()
        transactions["transaction"] = df.apply(to_transaction, axis=1)
        transactions["weight_vector"] = weight_vector

        return transactions

    def find_support(self, df, weight_vector):
        """
        Learning from the data which is a good support to analyze the input data.
        Args:
            df [pd.DataFrame]: first batch from input stream
        Yields: iterator
            batch_itemset [pd.DataFrame]: mined itemsets
            nw_count [int]: number of processed rows
            w_count [float]: sum of the target column
        """

        def support_found(batch_itemset):
            return len(batch_itemset) > self._max_itemest_per_batch

        if df is None or df.empty:
            if self._target:
                yield pd.DataFrame(columns=["w_count", "nw_count"]), 0, 0
                return

            yield pd.DataFrame(columns=["nw_count"]), 0, 0
            return

        transactions = self.get_transactions(df, weight_vector)

        # logarithmic progression of supports level to test out for the PopDelta's first batch
        supports = [88, 78, 69, 61, 53, 47, 42, 37, 35, 32, 28, 25, 22, 20, 17, 15, 13, 10, 7, 5, 3, 1]
        len_supports = len(supports)
        for ix, support in enumerate(supports):

            # incremental exploration
            upper_bound = max(500, int((ix / len_supports) * len(transactions)))
            partial_transactions = transactions[: upper_bound]
            partial_df = df[: upper_bound]
            partial_weight_vector = weight_vector[: upper_bound]

            batch_itemset = self.mine_batch(df=partial_df,
                                            weight_vector=partial_weight_vector,
                                            transactions=partial_transactions,
                                            support=support,
                                            max_len=self._max_len)

            nw_count, w_count = len(partial_transactions), sum(partial_weight_vector)
            yield batch_itemset, nw_count, w_count
            if support_found(batch_itemset):
                nw_count, w_count = len(transactions), sum(transactions["weight_vector"])
                batch_itemset = self.mine_batch(df=df,
                                                weight_vector=weight_vector,
                                                transactions=transactions,
                                                support=support,
                                                max_len=self._max_len)

                self._support = support
                yield batch_itemset, nw_count, w_count
                return

    def mine_batch(self, df, weight_vector, transactions, support, max_len):
        """
        Find Itemsets within the first batch in the input stream.
        If a weight is provided, then It will perform two parallel FIS search
        and lookups to fill the gaps beween missing itemsets.

        Returns:
            result [pd.DataFrame]: mined itemsets counts
        """

        def scale_weights(transactions):
            max_weight = transactions["weight_vector"].max()
            multiplier = (np.iinfo(np.int32).max / (len(transactions) * max_weight))
            transactions["weight_vector"] = (transactions["weight_vector"] * multiplier).round().astype(np.int32)
            return transactions

        transactions = transactions.groupby(by="transaction", as_index=False).sum()
        if self._target is not None:
            pos_transactions = transactions[transactions["weight_vector"] > 0]
            pos_itemsets = set()
            if len(pos_transactions) > 0:
                pos_transactions = scale_weights(pos_transactions)
                pos_result = self.fpgrowth_wrapper(pos_transactions, support, max_len, "w_count")
                pos_itemsets = set(pos_result.index.values)

            neg_transactions = transactions[transactions["weight_vector"] < 0]
            neg_itemsets = set()
            if len(neg_transactions) > 0:
                neg_transactions["weight_vector"] *= -1
                neg_transactions = scale_weights(neg_transactions)
                neg_result = self.fpgrowth_wrapper(neg_transactions, support, max_len, "w_count")
                neg_itemsets = set(neg_result.index.values)

            itemset = list(pos_itemsets.union(neg_itemsets))

            result = self.lookup_itemset(np.array(itemset, dtype=tuple), df,
                                         weight_vector=weight_vector, non_weighted=True,
                                         weighted=True)
        else:
            result = self.fpgrowth_wrapper(transactions, support, max_len, "nw_count")

        return result

    def fpgrowth_wrapper(self, transactions, support, max_len, colname):
        """
        pyFIM library wrapper

        Args:
            transactions (dict): transaction-weight map
            support (int): minimum support to mine
            max_len (int): max itemset driver length

        Returns:
            [np.array]: array of itemsets-count tuples
        """
        transactions = dict(transactions.values)
        transactions[("DUMMY_TRANSACTION",)] = 1  # dummy entry to avoid 100 support

        itemset = fpgrowth(transactions,
                           target="s",
                           supp=support,
                           zmax=max_len)
        result = pd.DataFrame(itemset, columns=["itemset", colname])
        result["itemset"] = result["itemset"].apply(lambda x: tuple(sorted(x)))
        result = result.set_index("itemset").astype({colname: float})

        if ("DUMMY_TRANSACTION",) in result.index:
            result.drop(index=("DUMMY_TRANSACTION",), inplace=True)

        return result

    def lookup_itemset(self, itemsets, df, weight_vector, non_weighted, weighted, query_step=1000):
        """
        Directly query batch to find itemsets frequencies.


        Args:
            itemsets (list): list of itemset to lookup on the dataframe df
            df (pd.DataFrame): Dataframe on which the lookup needs to be perfomed
            weight_vector (np.array): array to weight the transactions
            non_weighted (bool): is this a NWPopDelta ?
            query_step (int, optional): Number of itemsets to lookup at each iteration. Defaults to 1000.

        Returns:
            result [pd.DataFrame]:
        """
        # weighted = weight_vector is not None

        if non_weighted and weighted:
            result = pd.DataFrame(columns=["w_count", "nw_count"])
        elif non_weighted:
            result = pd.DataFrame(columns=["nw_count"])
        elif weighted:
            result = pd.DataFrame(columns=["w_count"])

        if df is None or df.empty:
            result["itemsets"] = itemsets
            result = result.set_index("itemsets")
            result = result.fillna(0)
            return result
        if itemsets.size == 0:
            return result

        df["zero"] = False

        colname_map = defaultdict(lambda: -1)
        colname_map.update({item: ix for ix, item in enumerate(df.columns)})

        lengths = np.vectorize(len)(itemsets)

        nw_lookups_count, w_lookups_count, new_index = [], [], []

        for lower_bound in range(0, len(itemsets), query_step):
            upper_bound = lower_bound + query_step
            step_itemsets = itemsets[lower_bound: upper_bound]
            step_lengths = lengths[lower_bound: upper_bound]
            for k in range(1, self._max_len + 1):
                equi_length_itemsets = step_itemsets[step_lengths == k]
                new_index += equi_length_itemsets.tolist()
                step_itemsets_ix = [[colname_map[item] for item in itemset] for itemset in equi_length_itemsets]

                if step_itemsets_ix:
                    binary_lookups_matrix = np.all(df.values[:, step_itemsets_ix], axis=2)

                    if non_weighted:
                        nw_lookups_count += list(np.sum(binary_lookups_matrix, axis=0).reshape(-1))
                    if weighted:
                        w_lookups_count += list(np.dot(weight_vector, binary_lookups_matrix).reshape(-1))

        if non_weighted:
            result["nw_count"] = nw_lookups_count
            result["nw_count"] = result["nw_count"].astype(int)
        if weighted:
            result["w_count"] = w_lookups_count
            result["w_count"] = result["w_count"].astype(float)
        result.index = new_index
        result.index.name = "itemset"

        return result

    def cumulate_result(self, old, new):
        """Update cumulative itemset result dataframe using the current batch results"""

        old = old.append(new).groupby(by="itemset", sort=False).sum()
        return old



# pylint: disable=W0612, W0613
"""
Support functions to augment the PopDelta results to support
all the different use cases.

Acronyms Glossary:
* NW = non-weighted (frequencies)
* W = weighted (frequencies weighted by the target columns)
* NWPopDelta1 = non-weighted PopDelta on 1 population (without target)
* WPopDelta1 = weighted PopDelta on 1 population (with target)
* NWPopDelta2 = non-weighted PopDelta on 2 population (without target)
* WPopDelta2 = weighted PopDelta on 2 population (with target)
"""

from copy import deepcopy
import itertools
import pandas as pd

from .settings import *  # noqa: F401, F403, pylint: disable=wildcard-import


def explode_itemsets(df):
    """
    Explodes input dataframe 3 itemset columns to six expanded columns.
    If an itemset has length smaller than 3, the entries in that column will be NA.

    From the form:
    itemsets1_bin-value1|itemsets2_bin-value2|itemsets3_bin-value3|metric1|...|metric-nth
    to:
    itemset1|bin-value1|itemset2|bin-value1|itemset1|bin-value1|metric1|...|metric-nth

    or more concretly, from
    age_2|weight_3|height_0|mean|...|impact
    to:
    age|2|weight|3|height|0|mean|...|impact

    Args:
        df ([pd.DataFrane]):  input dataframe which itemset column is going to be exploded
    """
    def explode(itemset):
        itemset = map(lambda item: item.rsplit("_", 1), itemset)
        return list(itertools.chain.from_iterable(itemset))

    exploded_colname = ["item1", "value1", "item2", "value2", "item3", "value3"]
    dummy = [0] * 6  # in case the max len itemset is < 3 this forces the df to 6 columns
    itemsets = [explode(itemset) for itemset in df.index.values]
    itemsets.append(dummy)
    exploded_df = pd.DataFrame(itemsets, columns=exploded_colname)[:-1]
    exploded_df.index = df.index
    exploded_df[df.columns] = df
    exploded_df[["item1", "item2", "item3"]] = exploded_df[["item1", "item2", "item3"]].astype(str)

    exploded_df[["value1", "value2", "value3"]] = exploded_df[["value1", "value2", "value3"]]#.astype(float) \
                                                                                             #.astype(dtype="Int64")
    return exploded_df


def format_nwpopdelta1(df, nw_sum_left, scaling_factors):
    """
    Add to the result dataframe the metrics related to NWPopDelta1,
    non-weighted, without target, PopDelta with 1 population.

    Given Fields:
    * nw_sum_left: non-weighted sum, aka number of rows, form left dataframe  (from fpgrowth)
    """

    inter_df = deepcopy(df)
    inter_df = inter_df.rename(columns={"nw_count": ABSOLUTE_LEFT}).astype(float)

    # scaling results on the whole dataset
    nw_sum_left *= scaling_factors[0]
    inter_df[[ABSOLUTE_LEFT]] *= scaling_factors[0]

    inter_df[[ABSOLUTE_LEFT]].fillna(0, inplace=True)
    inter_df[RELATIVE_LEFT] = inter_df[ABSOLUTE_LEFT] / nw_sum_left

    inter_df[LENGTH] = inter_df.index.map(len)
    inter_df = explode_itemsets(inter_df)
    result = inter_df.sort_values(by=ABSOLUTE_LEFT, ascending=False)

    return result


def format_wpopdelta1(df, nw_sum_left, w_sum_left, scaling_factors):
    """
    Add to the result dataframe the metrics related to WPopDelta1,
    weighted (with target) PopDelta with 1 population.

    Given Fields:
    * w_sum_left: non-weighted sum, aka number of rows, form left dataframe  (from fpgrowth)
    * nw_sum_left: non-weighted sum, aka number of rows, form left dataframe  (from lookup)
    """

    nw_sum_left = scaling_factors[0] * nw_sum_left
    w_sum_left = scaling_factors[0] * w_sum_left

    inter_df = deepcopy(df)
    inter_df = inter_df.rename(columns={"w_count": CUMULATIVE_LEFT,
                                        "nw_count": ABSOLUTE_LEFT}).astype(float)

    # scaling results on the whole dataset
    nw_sum_left *= scaling_factors[0]
    w_sum_left *= scaling_factors[0]
    inter_df[[CUMULATIVE_LEFT, ABSOLUTE_LEFT]] *= scaling_factors[0]

    inter_df[[CUMULATIVE_LEFT, ABSOLUTE_LEFT]].fillna(0, inplace=True)

    # compute metrics
    inter_df[RELATIVE_LEFT] = inter_df[ABSOLUTE_LEFT] / nw_sum_left
    inter_df[AVERAGE_LEFT] = inter_df[CUMULATIVE_LEFT] / inter_df[ABSOLUTE_LEFT]
    inter_df[CONTRIBUTION_LEFT] = inter_df[CUMULATIVE_LEFT] / w_sum_left
    inter_df[DIFFERENCE_GLOABL_AVERAGE] = inter_df[AVERAGE_LEFT] - (w_sum_left / nw_sum_left)

    inter_df[LENGTH] = inter_df.index.map(len)
    inter_df = explode_itemsets(inter_df)
    result = inter_df.sort_values(by=DIFFERENCE_GLOABL_AVERAGE, ascending=False)

    return result


def format_nwpopdelta2(df, nw_sum_left, nw_sum_right, scaling_factors):
    """
    Add to the result dataframe the metrics related to NWPopDelta2,
    non-weighted (without target) PopDelta with 2 population.

    Given Fields:
    * nw_sum_left: non-weighted sum, aka number of rows, form left dataframe (from fpgrowth)
    * nw_sum_right: non-weighted sum, aka number of rows, form right dataframe  (from fpgrowth)
    """

    inter_df = deepcopy(df)
    inter_df = df.rename(columns={"nw_count_left": ABSOLUTE_LEFT,
                                  "nw_count_right": ABSOLUTE_RIGHT}).astype(float)

    # scaling results on the whole dataset
    nw_sum_left *= scaling_factors[0]
    nw_sum_right *= scaling_factors[1]
    inter_df[ABSOLUTE_LEFT] *= scaling_factors[0]
    inter_df[ABSOLUTE_RIGHT] *= scaling_factors[1]

    inter_df[[ABSOLUTE_LEFT, ABSOLUTE_RIGHT]].fillna(0, inplace=True)

    # compute metrics
    inter_df[ABSOLUTE_DIFFERENCE] = inter_df[ABSOLUTE_LEFT] - inter_df[ABSOLUTE_RIGHT]
    inter_df[ABSOLUTE_RATIO_LR] = inter_df[ABSOLUTE_LEFT] / inter_df[ABSOLUTE_RIGHT]
    inter_df[ABSOLUTE_RATIO_RL] = inter_df[ABSOLUTE_RIGHT] / inter_df[ABSOLUTE_LEFT]

    inter_df[RELATIVE_LEFT] = inter_df[ABSOLUTE_LEFT] / nw_sum_left
    inter_df[RELATIVE_RIGHT] = inter_df[ABSOLUTE_RIGHT] / nw_sum_right
    inter_df[RELATIVE_DIFFERENCE] = inter_df[RELATIVE_LEFT] - inter_df[RELATIVE_RIGHT]
    inter_df[RELATIVE_RATIO_LR] = inter_df[RELATIVE_LEFT] / inter_df[RELATIVE_RIGHT]
    inter_df[RELATIVE_RATIO_RL] = inter_df[RELATIVE_RIGHT] / inter_df[RELATIVE_LEFT]

    inter_df = risk_ratio(inter_df, ABSOLUTE_LEFT, ABSOLUTE_RIGHT, nw_sum_left, nw_sum_right)

    inter_df[LENGTH] = inter_df.index.map(len)
    inter_df = explode_itemsets(inter_df)
    result = inter_df.sort_values(by=RISK_RATIO_LR, ascending=False)

    return result


def format_wpopdelta2(df, nw_sum_left, nw_sum_right, w_sum_left, w_sum_right, scaling_factors):
    """
    Add to the result dataframe the metrics related to WPopDelta2,
    weighted (with target) PopDelta with 2 population.

    Given Fields:
    * w_count_left: sum of the weight vector form left dataframe (from fpgrowth)
    * w_count_right: sum of the weight vector from right dataframe (from fpgrowth)
    * nw_count_left: non-weighted sum, aka number of rows, form left dataframe (from lookup)
    * nw_count_right: non-weighted sum, aka number of rows, form right dataframe (from lookup)
    """

    inter_df = deepcopy(df)
    inter_df = inter_df.rename(columns={"w_count_left": CUMULATIVE_LEFT,
                                        "nw_count_left": ABSOLUTE_LEFT,
                                        "w_count_right": CUMULATIVE_RIGHT,
                                        "nw_count_right": ABSOLUTE_RIGHT}).astype(float)

    # scaling results on the whole dataset
    nw_sum_left *= scaling_factors[0]
    nw_sum_right *= scaling_factors[1]
    w_sum_left *= scaling_factors[0]
    w_sum_right *= scaling_factors[1]
    inter_df[[ABSOLUTE_LEFT, CUMULATIVE_LEFT]] *= scaling_factors[0]
    inter_df[[ABSOLUTE_RIGHT, CUMULATIVE_RIGHT]] *= scaling_factors[1]

    inter_df[[ABSOLUTE_LEFT, ABSOLUTE_RIGHT, CUMULATIVE_LEFT, CUMULATIVE_RIGHT]].fillna(0, inplace=True)

    # compute metrics
    inter_df[ABSOLUTE_DIFFERENCE] = inter_df[ABSOLUTE_LEFT] - inter_df[ABSOLUTE_RIGHT]
    inter_df[ABSOLUTE_RATIO_LR] = inter_df[ABSOLUTE_LEFT] / inter_df[ABSOLUTE_RIGHT]
    inter_df[ABSOLUTE_RATIO_RL] = inter_df[ABSOLUTE_RIGHT] / inter_df[ABSOLUTE_LEFT]

    inter_df[RELATIVE_LEFT] = inter_df[ABSOLUTE_LEFT] / nw_sum_left
    inter_df[RELATIVE_RIGHT] = inter_df[ABSOLUTE_RIGHT] / nw_sum_right
    inter_df[RELATIVE_DIFFERENCE] = inter_df[RELATIVE_LEFT] - inter_df[RELATIVE_RIGHT]
    inter_df[RELATIVE_RATIO_LR] = inter_df[RELATIVE_LEFT] / inter_df[RELATIVE_RIGHT]
    inter_df[RELATIVE_RATIO_RL] = inter_df[RELATIVE_RIGHT] / inter_df[RELATIVE_LEFT]

    inter_df[CUMULATIVE_DIFFERENCE] = inter_df[CUMULATIVE_LEFT] - inter_df[CUMULATIVE_RIGHT]
    inter_df[CUMULATIVE_RATIO_LR] = inter_df[CUMULATIVE_LEFT] / inter_df[CUMULATIVE_RIGHT]
    inter_df[CUMULATIVE_RATIO_RL] = inter_df[CUMULATIVE_RIGHT] / inter_df[CUMULATIVE_LEFT]

    inter_df[AVERAGE_LEFT] = inter_df[CUMULATIVE_LEFT] / inter_df[ABSOLUTE_LEFT]
    inter_df[AVERAGE_RIGHT] = inter_df[CUMULATIVE_RIGHT] / inter_df[ABSOLUTE_RIGHT]
    inter_df[AVERAGE_DIFFERENCE] = inter_df[AVERAGE_LEFT] - inter_df[AVERAGE_RIGHT]
    inter_df[AVERAGE_RATIO_LR] = inter_df[AVERAGE_LEFT] / inter_df[AVERAGE_RIGHT]
    inter_df[AVERAGE_RATIO_RL] = inter_df[AVERAGE_RIGHT] / inter_df[AVERAGE_LEFT]

    inter_df[CONTRIBUTION_LEFT] = inter_df[CUMULATIVE_LEFT] / w_sum_left
    inter_df[CONTRIBUTION_RIGHT] = inter_df[CUMULATIVE_RIGHT] / w_sum_right
    inter_df[CONTRIBUTION_DIFFERENCE] = inter_df[CONTRIBUTION_LEFT] - inter_df[CONTRIBUTION_RIGHT]
    inter_df[CONTRIBUTION_RATIO_LR] = inter_df[CONTRIBUTION_LEFT] / inter_df[CONTRIBUTION_RIGHT]
    inter_df[CONTRIBUTION_RATIO_RL] = inter_df[CONTRIBUTION_RIGHT] / inter_df[CONTRIBUTION_LEFT]

    inter_df = risk_ratio(inter_df, CUMULATIVE_LEFT, CUMULATIVE_RIGHT, w_sum_left, w_sum_right)

    inter_df[LENGTH] = inter_df.index.map(len)
    inter_df = explode_itemsets(inter_df)
    result = inter_df.sort_values(by=RISK_RATIO_LR, ascending=False)

    return result


def risk_ratio(df, left_colname, right_colname, sum_left, sum_right):
    """Compute Risk Ratio for PopDelta2 (PopDelta 2 Populations)
    For reference: https://cs.stanford.edu/~matei/papers/2019/vldb_macrobase_diff.pdf (page 4)

    1) Prob that a row will be in the A relation vs. the B relation, if it has an attribute vs. if it does not
        * risk_ratio_a_b = P(A|Itemset) / P(A|~Itemset)

    2) Prob that a row will be in the B relation vs. the A relation, if it has an attribute vs. if it does not
        * risk_ratio_b_a = P(B|Itemset) / P(B|~Itemset)
    """

    # risk ratio a over b
    num1 = df[left_colname] / (df[left_colname] + df[right_colname])
    den1_num = sum_left - df[left_colname]
    den1_den = (sum_left - df[left_colname]) + (sum_right - df[right_colname])
    den1 = den1_num / den1_den
    df[RISK_RATIO_LR] = num1 / den1

    # risk ratio b over a
    num2 = df[right_colname] / (df[right_colname] + df[left_colname])
    den2_num = sum_right - df[right_colname]
    den2_den = (sum_right - df[right_colname]) + (sum_left - df[left_colname])
    den2 = den2_num / den2_den
    df[RISK_RATIO_RL] = num2 / den2

    # df["max_risk_ratio"] = np.max(df[["risk_ratio_a_b", "risk_ratio_b_a"]].values, axis=1)

    return df



# pylint: disable=c-extension-no-member, arguments-differ
"""Hot Encoding Operator"""

from typing import List
import pandas as pd
from pandas.api.types import is_numeric_dtype, is_datetime64_any_dtype, is_categorical_dtype
import numpy as np

class HotEncoder():
    """"Operator that implements One-hot-encoding"""

    def __init__(self, target, string_attributes, num_bins):  # pylint: disable=unused-argument
        """
        Args:
            description (dict): protobuff input message
        """
        self._target = target
        self._max_cardinality = 20
        self._string_attributes = set(string_attributes)
        self._columns_to_drop = []
        self._bins = {}
        self._num_bins = num_bins
        self._columns_to_bin = []
        self._binary_columns = []
        self._non_binary_columns = []
        self._initialized = False
        self._warnings = []
        self._dropped_rows = 0

    def reset(self):
        self._columns_to_drop = []
        self._binary_columns = []
        self._non_binary_columns = []
        self._initialized = False

    def detect_columns_issues(self, df):
        """
        detect high cardinality, constant and binary columns
        """
        for col_name, column_values in df.items():
            unique_values = set(column_values.value_counts(dropna=False).index)
            num_unique = len(unique_values)
            # high cardinal columns
            if num_unique > self._max_cardinality:
                if is_numeric_dtype(df[col_name]) or is_datetime64_any_dtype(df[col_name]):
                    self._columns_to_bin.append(col_name)
                else:
                    self._columns_to_drop.append(col_name)
                    msg = f'Column {col_name} dropped because it has a high cardinality (>100).'
                    self._warnings.append(msg)
            if is_categorical_dtype(df[col_name]):
                self._columns_to_drop.append(col_name)
                msg = f'Column {col_name} dropped because it is a categorical column.'
                self._warnings.append(msg)
            # constant columns
            elif num_unique == 1:
                self._columns_to_drop.append(col_name)
            # null columns
            elif column_values.isna().all():
                self._columns_to_drop.append(col_name)
            # boolean columns
            elif (unique_values == {0, 1}) and (unique_values == {0.0, 1.0}) and (unique_values == {True, False}) and (col_name not in self._string_attributes):
                self._binary_columns.append(col_name)
        self._non_binary_columns = list(set(df.columns) - set(self._columns_to_drop) - set(self._binary_columns))

    # pylint: disable=cell-var-from-loop
    def process_batch(self, batches: List[pd.DataFrame]) -> pd.DataFrame:
        """
        Hot-encode input batch

        Args:
            batches (List[pd.DataFrame]): input batch

        Returns:
            pd.DataFrame: hot-encoded dataframe
        """

        def get_int_bin_val(col_name):
            # convert from age_1.0 to age_1

            att, bin_val = col_name.rsplit("_", 1)
            if bin_val == "nan":
                return col_name
            return "_".join([att, bin_val[:-2]])

        df = batches[0]
        # display(df)
        # print(df["age"])

        # remove target variable
        if self._target is not None:
            nrows = len(df)
            df.dropna(subset=[self._target], inplace=True)  # dropping rows which are nan on target
            self._dropped_rows += nrows - len(df)
            if self._dropped_rows != 0:
                msg = "Some rows dropped because they contained NaN values in the target column"
                self._warnings.append(msg)
            self._target_values = df[self._target].values
            df.drop(self._target, axis=1, inplace=True)

        if not self._initialized:
            self.detect_columns_issues(df)

        # drop high cardinality, constant or problematic columns
        df = df.drop(self._columns_to_drop, axis=1)

        # bin numerical columns
        for col_name in self._columns_to_bin:
            # first iteration - create bins
            if not self._initialized:
                df[col_name], labels = pd.cut(df[col_name], bins=self._num_bins, labels=None, retbins=True)
                self._bins[col_name] = labels
            else:
                # second iteration - use bins
                df[col_name] = pd.cut(df[col_name], bins=self._bins[col_name], labels=False)
            df[col_name] = df[col_name].astype(str)
        self._initialized = True

        if df.empty:
            print("No compatible attribute to analyze in the input dataframes. Please select other attributes.")


        # drop first element of binary columns
        df.fillna("nan", inplace=True)
        df = pd.get_dummies(df, columns=self._binary_columns, drop_first=True, dummy_na=True)
        df = pd.get_dummies(df, columns=self._non_binary_columns, dummy_na=True)

        df.columns = [get_int_bin_val(col_name) for col_name in df.columns if col_name not in self._binary_columns]

        # add back the target column
        if self._target is not None:
            df[self._target] = self._target_values

        return df, self._warnings


__all__ = ['HotEncoder']



"""PopDelta related operator global variables definition"""

ABSOLUTE_LEFT = "LEFT Presence (#)"
ABSOLUTE_RIGHT = "RIGHT Presence (#)"
ABSOLUTE_DIFFERENCE = "Presence Difference (Δ)"
ABSOLUTE_RATIO_LR = "Presence Ratio LEFT/RIGHT"
ABSOLUTE_RATIO_RL = "Presence Ratio RIGHT/LEFT"

RELATIVE_LEFT = "LEFT Prevalence (%)"
RELATIVE_RIGHT = "RIGHT Prevalence (%)"
RELATIVE_DIFFERENCE = "Prevalence Difference (Δ)"
RELATIVE_RATIO_LR = "Prevalence Ratio LEFT/RIGHT"
RELATIVE_RATIO_RL = "Prevalence Ratio RIGHT/LEFT"

CUMULATIVE_LEFT = "LEFT TARGET (#)"
CUMULATIVE_RIGHT = "RIGHT TARGET (#)"
CUMULATIVE_DIFFERENCE = "TARGET Difference (Δ)"
CUMULATIVE_RATIO_LR = "Target Ratio LEFT/RIGHT"
CUMULATIVE_RATIO_RL = "Target Ratio RIGHT/LEFT"

CONTRIBUTION_LEFT = "LEFT TARGET Share (%)"
CONTRIBUTION_RIGHT = "RIGHT TARGET Share (%)"
CONTRIBUTION_DIFFERENCE = "TARGET Share Difference (Δ)"
CONTRIBUTION_RATIO_LR = "Share Ratio LEFT/RIGHT"
CONTRIBUTION_RATIO_RL = "Share Ratio RIGHT/LEFT"

AVERAGE_LEFT = "LEFT Average TARGET"
AVERAGE_RIGHT = "RIGHT Average TARGET"
AVERAGE_DIFFERENCE = "Average TARGET Difference (Δ)"
DIFFERENCE_GLOABL_AVERAGE = "Difference with Global Average (Δ)"
AVERAGE_RATIO_LR = "Average Ratio LEFT/RIGHT"
AVERAGE_RATIO_RL = "Average Ratio RIGHT/LEFT"


RISK_RATIO_LR = "Risk Ratio LEFT over RIGHT"
RISK_RATIO_RL = "Risk Ratio RIGHT over LEFT"


LENGTH = "Length"



# pylint: disable=c-extension-no-member, arguments-differ, W0401, W0614
"""The PopDelta operator allows to automatically discover fundamental patterns/drivers in your data.
You can use it to characterize a single population (market basket analysis), or to find the differences
between two populations"""

from typing import List

import pandas as pd
import warnings

from .pop_delta1 import PopDelta1
from .pop_delta2 import PopDelta2
from .hot_encoder import HotEncoder

class PopDelta():
    """
    The class implementing PopDelta. It first bins and hot-encodes the data and then
    based on the type of PopDelta runs fpgrowth on 1 or 2 streams.
    """
    def __init__(self, target, num_bins, string_attributes):  # pylint: disable=unused-argument
        self.target = target if target != "" else None
        self.has_target = self.target is not None
        self.max_len = 3
        self._pop_delta = None
        self._encoder = None
        self._encoded_left = None
        self._encoded_right = None
        self._batch = None
        self._DEFAULT_CHUNK_SIZE = 10**4
        self._encoder = HotEncoder(self.target, string_attributes, num_bins)

    def _select_use_case(self, batches):
        self.max_len = min(self.max_len, len(batches[0].columns))
        if len(batches) == 1:
            pop_delta = PopDelta1(self)
        else:
            # len(batches) == 2
            pop_delta = PopDelta2(self)
        return pop_delta

    def _select_max_len(self, batches):
        """
        Select Itemset max length based on the default parameter and
        the input dataframe.
        """
        if len(batches) == 1:
            max_len = min(len(batches[0].columns), self.max_len)
        elif batches[0] is not None and batches[1] is not None:
            max_len = min(len(batches[0].columns), self.max_len)
        elif batches[0] is not None:
            max_len = min(len(batches[0].columns), self.max_len)
        else:  # batches[1] is not None:
            max_len = min(len(batches[1].columns), self.max_len)

        return max_len

    def _hot_encode(self, batches):
        """
        Hot-Encodes the binned input dataframe's columns to make it
        compatible with itemset mining. If two dataframes are given
        as input,  they are first vertically appended, then hot-encoded,
        and finally vertically split again.

        This ensures consistency of the columns.

        Args:
            batches List[pd.DataFrame]: Binned Dataframes

        Returns:
            batches List[pd.DataFrame]: Hot-Encoded Dataframes
        """


        if len(batches) == 1:
            # single batch
            encoded_batch, hot_encoder_warnings = self._encoder.process_batch(batches)
            batches = [encoded_batch]
        elif batches[0] is not None and batches[1] is not None:
            # both batches are good
            stacked_batch = [batches[0].append(batches[1])]
            encoded_batch, hot_encoder_warnings = self._encoder.process_batch(stacked_batch)
            batches = [encoded_batch[:len(batches[0])], encoded_batch[len(batches[0]):]]
        elif batches[0] is not None:
            # left batch is good, right batch is empty
            left_batch, hot_encoder_warnings = self._encoder.process_batch([batches[0]])
            batches = [left_batch, pd.DataFrame(columns=left_batch.columns)]
        else:
            # right batch is good, left batch is empty
            right_batch, hot_encoder_warnings = self._encoder.process_batch([batches[1]])
            batches = [pd.DataFrame(columns=right_batch.columns), right_batch]

        return batches

    def _one_chunks_generator(self, df1, chunk_size):
        df1 = df1.copy(deep=True).sample(frac=1)
        num_chunks_df1 = len(df1) // chunk_size + (1 if len(df1) % chunk_size != 0 else 0)

        for i in range(num_chunks_df1):
            start = i * chunk_size
            end = (i + 1) * chunk_size
            self.INPUT_SCALING_FACTORS = [end / len(df1)]

            chunk1 = df1.iloc[start:end] if i < num_chunks_df1 else None

            yield chunk1

    def _two_chunks_generator(self, df1, df2, chunk_size):
        df1 = df1.sample(frac=1).copy(deep=True)
        df2 = df2.sample(frac=1).copy(deep=True)

        num_chunks_df1 = len(df1) // chunk_size + (1 if len(df1) % chunk_size != 0 else 0)
        num_chunks_df2 = len(df2) // chunk_size + (1 if len(df2) % chunk_size != 0 else 0)

        for i in range(max(num_chunks_df1, num_chunks_df2)):
            start = i * chunk_size
            end = (i + 1) * chunk_size
            self.INPUT_SCALING_FACTORS = [end / len(df1), end / len(df2)]

            chunk1 = df1.iloc[start:end] if i < num_chunks_df1 else None
            chunk2 = df2.iloc[start:end] if i < num_chunks_df2 else None

            yield [chunk1, chunk2]



    def process_batch(self, batches: List[pd.DataFrame], progressive=False) -> pd.DataFrame:
        warnings.filterwarnings('ignore')

        if len(batches) == 1 & progressive == True:
            chunks_generator = self._one_chunks_generator(df1=batches[0], chunk_size=self._DEFAULT_CHUNK_SIZE)
        elif len(batches) == 2 & progressive == True:
            chunks_generator = self._two_chunks_generator(df1=batches[0], df2=batches[1], chunk_size=self._DEFAULT_CHUNK_SIZE)
        else:
            chunks_generator = [[b.copy(deep=True) for b in batches]]
            self.INPUT_SCALING_FACTORS = [1. for b in batches]



        for chunks in chunks_generator:
            encoded_batches = self._hot_encode(chunks)
            scaling_factors = self.INPUT_SCALING_FACTORS
            if self._pop_delta is None:
                self.max_len = self._select_max_len(chunks)
                self._pop_delta = self._select_use_case(encoded_batches)

            for result in self._pop_delta.process_batch(encoded_batches, scaling_factors):
                yield result


__all__ = ['PopDelta']
